var fs = require("fs"), util = require("util"), javascriptParser = require("uglify-js").parser, ErrCodes = require("./err_codes");

exports.construct = function(fileName, ownerFilename, callback) {
    fs.readFile(fileName, function(readErr, data) {
        if (readErr) {
            callback({
                code: ownerFilename ? ErrCodes.FAILED_LOAD_REQUIRE : ErrCodes.READ_FILE_FAILED,
                filename: fileName,
                ownerFilename: ownerFilename
            });
            return;
        }
        var ast = null, srcCode = data.toString().trim();
        srcCode = srcCode.replace(/\r\n?|[\n\u2028\u2029]/g, "\n").replace(/^\uFEFF/, "");
        try {
            ast = javascriptParser.parse(srcCode, false, true);
        } catch (parseErr) {
            callback({
                code: ErrCodes.JAVASCRIPT_PARSING_FAILED,
                filename: fileName,
                parserErr: parseErr
            });
            return;
        }
        callback(null, ast, srcCode);
    });
};

var getEntryName = exports.getEntryName = function(entry) {
    return typeof entry[0] === "object" ? entry[0].name : entry[0];
};

exports.isPathMatch = function(expectedPath, actualPath, ensureLast) {
    if (expectedPath.length !== actualPath.length - 1) return false;
    if (ensureLast && !actualPath[1][0].last) return false;
    for (var i = 0; i < expectedPath.length; i++) {
        var hasFlag = util.isArray(expectedPath[i]), name = hasFlag ? expectedPath[i][0] : expectedPath[i];
        if (getEntryName(actualPath[i]) !== name || hasFlag && actualPath[i][1] !== expectedPath[i][1]) return false;
    }
    return true;
};

exports.getAncestorByName = function(name, currentAstPath) {
    if (currentAstPath.length > 1) {
        for (var i = currentAstPath.length - 2; i >= 0; i--) {
            if (getEntryName(currentAstPath[i]) === name) return currentAstPath[i];
        }
    }
    return null;
};

exports.getCurrentSrcLineNum = function(currentAst) {
    for (var i = currentAst.length - 1; i >= 0; i--) {
        if (currentAst[i]) {
            if (currentAst[i].start) return currentAst[i].start.line;
            if (currentAst[i][0] && currentAst[i][0].start) return currentAst[i][0].start.line;
        }
    }
    return 0;
};

exports.getRemainderAst = function(ast) {
    var astBranches = [];
    ast[1].forEach(function(statement) {
        if (!statement[0].remove) astBranches.push(statement);
    });
    return astBranches.length ? [ "toplevel", astBranches ] : null;
};

exports.getSrcCodePosFromPath = function(astPath) {
    return exports.getSrcCodePosFromEntry(astPath[astPath.length - 1]);
};

exports.getSrcCodePosFromEntry = function(astEntry) {
    var pos = {
        start: 0,
        end: 0
    };
    if (astEntry && astEntry[0] && astEntry[0].start) {
        pos.start = astEntry[0].start.pos;
        pos.end = astEntry[0].end.endpos;
    }
    return pos;
};